/********************************************************************************
 * Copyright (c) 2023 CEA-List
 *
 * This program and the accompanying materials are made available under the
 * terms of the Eclipse Public License 2.0 which is available at
 * http://www.eclipse.org/legal/epl-2.0.
 *
 * SPDX-License-Identifier: EPL-2.0
 *
 ********************************************************************************/

#ifndef AIDGE_CORE_OPERATOR_OPERATORTENSOR_H_
#define AIDGE_CORE_OPERATOR_OPERATORTENSOR_H_

#include <memory>
#include <string>
#include <vector>

#include "aidge/backend/OperatorImpl.hpp"
#include "aidge/data/Tensor.hpp"
#include "aidge/utils/Types.h"
#include "aidge/operator/Operator.hpp"

namespace Aidge {

class OperatorTensor : public Operator {
/* TODO: Add an attribute specifying the type of Data used by the Operator.
 * The same way ``Type`` attribute specifies the type of Operator. Hence this
 * attribute could be checked in the forwardDims function to assert Operators
 * being used work with Tensors and cast them to OpertorTensor instead of
 * Operator.
 */
/* TODO: Maybe change type attribute of Data object by an enum instead of an
 * array of char. Faster comparisons.
 */
protected:
    std::vector<std::shared_ptr<Tensor>*> mInputs;
    std::vector<std::shared_ptr<Tensor>> mOutputs;

public:
    OperatorTensor(const char* type, const IOIndex_t nbData, const IOIndex_t nbAttr, const IOIndex_t nbOut)
    : Operator(type, nbData, nbAttr, nbOut),
      mInputs(std::vector<std::shared_ptr<Tensor>*>(nbData + nbAttr, nullptr)),
      mOutputs(std::vector<std::shared_ptr<Tensor>>(nbOut))
    {
        for (std::size_t i = 0; i < static_cast<std::size_t>(nbOut); ++i) {
            mOutputs[i] = std::make_shared<Tensor>();
        }
    }

public:
    ///////////////////////////////////////////////////
    virtual void associateInput(const IOIndex_t inputIdx, const std::shared_ptr<Data>* data) override;
    ///////////////////////////////////////////////////

    ///////////////////////////////////////////////////
    // Tensor access
    // input management
    std::shared_ptr<Tensor> getInput(const IOIndex_t inputIdx) const;
    Tensor& input(const IOIndex_t inputIdx) const;
    std::shared_ptr<Data> getRawInput(const IOIndex_t inputIdx) const override final;

    //output management
    std::shared_ptr<Tensor> getOutput(const IOIndex_t outputIdx) const;
    Tensor& output(const IOIndex_t outputIdx) const;
    std::shared_ptr<Data> getRawOutput(const IOIndex_t outputIdx) const override final;
    ///////////////////////////////////////////////////

    ///////////////////////////////////////////////////
    // Tensor dimensions
    virtual void computeOutputDims() = 0;
    virtual bool outputDimsForwarded() const;
    ///////////////////////////////////////////////////

    virtual void setDataType(const DataType& dataType) const;

};
} // namespace Aidge

#endif // AIDGE_CORE_OPERATOR_OPERATORTENSOR_H_